#!/usr/bin/env python
#-*- coding: utf-8 -*-
#pylint: disable=
"""
File       : processor.py
Author     : Valentin Kuznetsov <vkuznet AT gmail dot com>
Description: Module for Processors used only by EDD experiments
"""

# system modules
from json import dumps

# third party modules
import numpy as np

# local modules
from CHAP.processor import Processor


class DiffractionVolumeLengthProcessor(Processor):
    """A Processor using a steel foil raster scan to calculate the
    length of the diffraction volume for an EDD setup.
    """

    def process(self, data, interactive=False):
        """Return calculated value of the DV length.

        :param data: input configuration for the raw scan data & DVL
            calculation procedure.
        :type data: list[PipelineData]
        :param dvl_model: method to use for calculating DVL. Choices:
            one of three acceptable scalars which will be multiplied
            by the standard deviation of a gauusian fit to the raster
            scan data, or "manual" (in which case the user is
            presented with a plot of the fit and unfit data, and they
            select the accepatble DVL by eye).
        :type dvl_model: Literal[1.0, 1.75, 2.0, "manual"]
        :param interactive: allow for user interactions, defaults to
            False
        :type interactive: bool, optional
        :return: complete DVL configuraiton dictionary
        :rtype: dict
        """

        dvl_config = self.get_config(
            data, 'edd.models.DiffractionVolumeLengthConfig')
        dvl = self.measure_dvl(dvl_config, interactive=interactive)
        dvl_config.dvl_measured = dvl

        return dvl_config.dict()

    def measure_dvl(self, dvl_config, interactive=False):
        """Return a measured value for the length of the diffraction
        volume. Use the iron foil raster scan data provided in
        `dvl_config` and fit a gaussian to the sum of all MCA channel
        counts vs scanned motor position in the raster scan. The
        computed diffraction volume length is approximately equal to
        the standard deviation of the fitted peak.

        :param dvl_config: configuration for the DVL calculation
            procedure
        :type dvl_config: DiffractionVolumeLengthConfig
        :param dvl_model: method to use for calculating DVL. Choices:
            one of three acceptable scalars which will be multiplied
            by the standard deviation of a gauusian fit to the raster
            scan data, or "manual" (in which case the user is
            presented with a plot of the fit and unfit data, and they
            select the accepatble DVL by eye).
        :type dvl_model: Literal[1.0, 1.75, 2.0, "manual"]
        :param interactive: allow for user interactions, defaults to
            False
        :type interactive: bool, optional
        :return: calculated diffraction volume length
        :rtype: float
        """

        from CHAP.utils.fit import Fit
        from CHAP.utils.general import draw_mask_1d

        # Get raw MCA data from raster scan
        mca_data = dvl_config.mca_data()

        # Interactively set mask, if needed & possible.
        if dvl_config.include_bin_ranges is None:
            if interactive:
                mask, include_bin_ranges = draw_mask_1d(
                    np.sum(mca_data, axis=0),
                    xdata = np.arange(dvl_config.num_bins),
                    label='sum of MCA spectra over all scan points',
                    title='Click and drag to select ranges of MCA data to\n'
                    + 'include when measuring the diffraction volume length.',
                    xlabel='MCA channel (index)',
                    ylabel='MCA intensity (counts)'
                )
                dvl_config.include_bin_ranges = include_bin_ranges
            else:
                raise ValueError(
                    'No value provided for include_bin_ranges. '
                    + 'Provide them in the Diffraction Volume Length '
                    + 'Measurement Configuration, or re-run the pipeline '
                    + 'with the --interactive flag.')

        # Reduce the raw MCA data in 3 ways:
        # 1) sum of intensities in all detector bins
        # 2) max of intensities in detector bins after mask is applied
        # 3) sum of intensities in detector bins after mask is applied
        unmasked_sum = np.sum(mca_data, axis=1)
        mask = dvl_config.mca_mask()
        masked_mca_data = np.empty(
            (mca_data.shape[0], *mca_data[0][mask].shape))
        for i in range(mca_data.shape[0]):
            masked_mca_data[i] = mca_data[i][mask]
        masked_max = np.amax(masked_mca_data, axis=1)
        masked_sum = np.sum(masked_mca_data, axis=1)

        # Find the motor position corresponding roughly to the center
        # of the diffraction volume
        motor_vals = dvl_config.motor_vals
        scan_center = np.sum(motor_vals * masked_sum) / np.sum(masked_sum)
        x = motor_vals - scan_center

        # "Normalize" the masked summed data and fit a gaussian to it
        y = (masked_sum - min(masked_sum)) / max(masked_sum)
        fit = Fit.fit_data(y, 'gaussian', x=x, normalize=False)

        # Calculate / manually select diffraction volume length
        dvl = fit.best_values['sigma'] * dvl_config.sigma_to_dvl_factor
        if interactive:
            from CHAP.utils.general import input_yesno
            manual_dvl = input_yesno(
                'Indicate the diffraction volume length manually? (y/n)',
                default='y')
            if manual_dvl:
                dvl_config.measurement_mode = 'manual'
                mask, dvl_bounds = draw_mask_1d(
                    y, xdata=x,
                    label='total (masked, normalized)',
                    ref_data=[
                        ((x, fit.best_fit),
                         {'label': 'gaussian fit'}),
                        ((x, masked_max / max(masked_max)),
                         {'label': 'maximum (masked, normalized)'}),
                        ((x, unmasked_sum / max(unmasked_sum)),
                         {'label': 'total (unmasked, normalized)'})
                    ],
                    num_index_ranges_max=1,
                    title=('Click and drag to indicate the\n'
                           + 'boundary of the diffraction volume'),
                    xlabel=(dvl_config.motor_mne
                            + ' (offset from scan "center")'),
                    ylabel='MCA intensity (counts)')
                dvl_bounds = dvl_bounds[0]
                dvl = abs(x[dvl_bounds[1]] - x[dvl_bounds[0]])

        return dvl

class MCACeriaCalibrationProcessor(Processor):
    """A Processor using a CeO2 scan to obtain tuned values for the
    bragg diffraction angle and linear correction parameters for MCA
    channel energies for an EDD experimental setup.
    """

    def process(self, data, interactive=False):
        """Return tuned values for 2&theta and linear correction
        parameters for the MCA channel energies.

        :param data: input configuration for the raw data & tuning
            procedure
        :type data: list[dict[str,object]]
        :param interactive: allow for user interactions, defaults to
            False
        :type interactive: bool, optional
        :return: original configuration dictionary with tuned values
            added
        :rtype: dict[str,float]
        """

        calibration_config = self.get_config(
            data, 'edd.models.MCACeriaCalibrationConfig')

        tth, slope, intercept = self.calibrate(calibration_config,
                                               interactive=interactive)

        calibration_config.tth_calibrated = tth
        calibration_config.slope_calibrated = slope
        calibration_config.intercept_calibrated = intercept

        return calibration_config.dict()

    def calibrate(self, calibration_config, interactive=False):
        """Iteratively calibrate 2&theta by fitting selected peaks of
        an MCA spectrum until the computed strain is sufficiently
        small. Use the fitted peak locations to determine linear
        correction parameters for the MCA's channel energies.

        :param calibration_config: object configuring the CeO2
            calibration procedure
        :type calibration_config: MCACeriaCalibrationConfig
        :param interactive: allow for user interactions, defaults to
            False
        :type interactive: bool, optional
        :return: calibrated values of 2&theta and linear correction
            parameters for MCA channel energies : tth, slope,
            intercept
        :rtype: float, float, float
        """
        # third party modules
        from scipy.constants import physical_constants

        # local modules
        from CHAP.utils.fit import Fit, FitMultipeak

        # We'll work in keV and A, not eV and m.
        hc = 1e7 * physical_constants['Planck constant in eV/Hz'][0] \
            * physical_constants['speed of light in vacuum'][0]

        # Collect raw MCA data of interest
        mca_data = calibration_config.mca_data()
        mca_bin_energies = np.arange(0, calibration_config.num_bins) \
            * (calibration_config.max_energy_kev/calibration_config.num_bins)

        # Mask out the corrected MCA data for fitting
        if calibration_config.include_bin_ranges is None:
            if interactive:
                from CHAP.utils.general import draw_mask_1d
                mask, include_bin_ranges = draw_mask_1d(
                    mca_data,
                    xdata=mca_bin_energies,
                    title='Click and drag to select ranges of Ceria'
                           +' calibration data to include',
                    xlabel='MCA channel energy (keV)',
                    ylabel='MCA intensity (counts)')
                calibration_config.include_bin_ranges = include_bin_ranges
            else:
                raise ValueError(
                    'No value provided for include_bin_ranges. '
                    'Provide them in the MCA Ceria Calibration Configuration, '
                    'or re-run the pipeline with the --interactive flag.')
        mca_mask = calibration_config.mca_mask()
        fit_mca_energies = mca_bin_energies[mca_mask]
        fit_mca_intensities = mca_data[mca_mask]

        # Correct raw MCA data for variable flux at different energies
        flux_correct = \
            calibration_config.flux_correction_interpolation_function()
        mca_intensity_weights = flux_correct(fit_mca_energies)
        fit_mca_intensities = fit_mca_intensities/mca_intensity_weights

        # Get the HKLs and lattice spacings that will be used for
        # fitting
        tth = calibration_config.tth_initial_guess
        if calibration_config.fit_hkls is None:
            if interactive:
                from CHAP.utils.general import select_peaks
                hkls, ds = calibration_config.unique_ds()
                peak_locations = hc / (2. * ds * np.sin(0.5 * np.radians(tth)))
                selected_peaks = select_peaks(
                    mca_data, mca_bin_energies, peak_locations,
                    mask=mca_mask)
                fit_hkls = [np.where(peak_locations == peak)[0][0]
                            for peak in selected_peaks]
                calibration_config.fit_hkls = fit_hkls
            else:
                raise ValueError(
                    'No value provided for fit_hkls. Provide them in '
                    'the MCA Ceria Calibration Configuration, or re-run '
                    'the pipeline with the --interactive flag.')
        fit_hkls, fit_ds = calibration_config.fit_ds()
        c_1 = fit_hkls[:,0]**2 + fit_hkls[:,1]**2 + fit_hkls[:,2]**2

        for iter_i in range(calibration_config.max_iter):

            # Perform the uniform fit first

            # Get expected peak energy locations for this iteration's
            # starting value of tth
            fit_lambda = 2.0*fit_ds*np.sin(0.5*np.radians(tth))
            fit_E0 = hc / fit_lambda

            # Run the uniform fit
            best_fit, residual, best_values, best_errors, redchi, success = \
                FitMultipeak.fit_multipeak(
                    fit_mca_intensities,
                    fit_E0,
                    x=fit_mca_energies,
                    fit_type='uniform',
                    plot=False)

            # Extract values of interest from the best values for the
            # uniform fit parameters
            uniform_fit_centers = [
                best_values[f'peak{i+1}_center']
                for i in range(len(calibration_config.fit_hkls))]
            # uniform_a = best_values['scale_factor']
            # uniform_strain = np.log(
            #     (uniform_a
            #      / calibration_config.lattice_parameter_angstrom))
            # uniform_tth = tth * (1.0 + uniform_strain)
            # uniform_rel_rms_error = (np.linalg.norm(residual)
            #                          / np.linalg.norm(fit_mca_intensities))

            # Next, perform the unconstrained fit

            # Use the peak locations found in the uniform fit as the
            # initial guesses for peak locations in the unconstrained
            # fit
            best_fit, residual, best_values, best_errors, redchi, success = \
                FitMultipeak.fit_multipeak(
                    fit_mca_intensities,
                    uniform_fit_centers,
                    x=fit_mca_energies,
                    fit_type='unconstrained',
                    plot=False)

            # Extract values of interest from the best values for the
            # unconstrained fit parameters
            unconstrained_fit_centers = np.array(
                [best_values[f'peak{i+1}_center']
                 for i in range(len(calibration_config.fit_hkls))])
            unconstrained_a = 0.5*hc*np.sqrt(c_1) \
                / (unconstrained_fit_centers*abs(np.sin(0.5*np.radians(tth))))
            unconstrained_strains = np.log(
                unconstrained_a/calibration_config.lattice_parameter_angstrom)
            unconstrained_strain = np.mean(unconstrained_strains)
            unconstrained_tth = tth * (1.0+unconstrained_strain)
            unconstrained_rel_rms_error = (
                np.linalg.norm(residual)/np.linalg.norm(fit_mca_intensities))

            # Update tth for the next iteration of tuning
            prev_tth = tth
            tth = unconstrained_tth

            # Stop tuning tth at this iteration if differences are
            # small enough
            if abs(tth - prev_tth) < calibration_config.tune_tth_tol:
                break

        # Fit line to expected / computed peak locations from the last
        # unconstrained fit.
        fit = Fit.fit_data(
            fit_E0,
            'linear',
            x=unconstrained_fit_centers,
            nan_policy='omit')
        slope = fit.best_values['slope']
        intercept = fit.best_values['intercept']

        return float(tth), float(slope), float(intercept)


class MCADataProcessor(Processor):
    """A Processor to return data from an MCA, restuctured to
    incorporate the shape & metadata associated with a map
    configuration to which the MCA data belongs, and linearly
    transformed according to the results of a ceria calibration.
    """

    def process(self, data):
        """Process configurations for a map and MCA detector(s), and
        return the calibrated MCA data collected over the map.

        :param data: input map configuration and results of ceria
            calibration
        :type data: list[dict[str,object]]
        :return: calibrated and flux-corrected MCA data
        :rtype: nexusformat.nexus.NXentry
        """

        map_config = self.get_config(
            data, 'common.models.map.MapConfig')
        calibration_config = self.get_config(
            data, 'edd.models.MCACeriaCalibrationConfig')
        nxroot = self.get_nxroot(map_config, calibration_config)

        return nxroot

    def get_nxroot(self, map_config, calibration_config):
        """Get a map of the MCA data collected by the scans in
        `map_config`. The MCA data will be calibrated and
        flux-corrected according to the parameters included in
        `calibration_config`. The data will be returned along with
        relevant metadata in the form of a NeXus structure.

        :param map_config: the map configuration
        :type map_config: MapConfig
        :param calibration_config: the calibration configuration
        :type calibration_config: MCACeriaCalibrationConfig
        :return: a map of the calibrated and flux-corrected MCA data
        :rtype: nexusformat.nexus.NXroot
        """
        # third party modules
        from nexusformat.nexus import (NXdata,
                                       NXdetector,
                                       NXinstrument,
                                       NXroot)

        # local modules
        from CHAP.common import MapProcessor

        nxroot = NXroot()

        nxroot[map_config.title] = MapProcessor.get_nxentry(map_config)
        nxentry = nxroot[map_config.title]

        nxentry.instrument = NXinstrument()
        nxentry.instrument.detector = NXdetector()
        nxentry.instrument.detector.calibration_configuration = dumps(
            calibration_config.dict())

        nxentry.instrument.detector.data = NXdata()
        nxdata = nxentry.instrument.detector.data
        nxdata.raw = np.empty((*map_config.shape, calibration_config.num_bins))
        nxdata.raw.attrs['units'] = 'counts'
        nxdata.channel_energy = calibration_config.slope_calibrated \
            * np.arange(0, calibration_config.num_bins) \
            * (calibration_config.max_energy_kev/calibration_config.num_bins) \
            + calibration_config.intercept_calibrated
        nxdata.channel_energy.attrs['units'] = 'keV'

        for scans in map_config.spec_scans:
            for scan_number in scans.scan_numbers:
                scanparser = scans.get_scanparser(scan_number)
                for scan_step_index in range(scanparser.spec_scan_npts):
                    map_index = scans.get_index(
                        scan_number,
                        scan_step_index,
                        map_config)
                    nxdata.raw[map_index] = scanparser.get_detector_data(
                        calibration_config.detector_name,
                        scan_step_index)

        nxentry.data.makelink(
            nxdata.raw,
            name=calibration_config.detector_name)
        nxentry.data.makelink(
            nxdata.channel_energy,
            name=f'{calibration_config.detector_name}_channel_energy')
        if isinstance(nxentry.data.attrs['axes'], str):
            nxentry.data.attrs['axes'] = [
                nxentry.data.attrs['axes'],
                f'{calibration_config.detector_name}_channel_energy']
        else:
            nxentry.data.attrs['axes'] += [
                f'{calibration_config.detector_name}_channel_energy']
        nxentry.data.attrs['signal'] = calibration_config.detector_name

        return nxroot


if __name__ == '__main__':
    # local modules
    from CHAP.processor import main

    main()



# third party modules
import numpy as np
from pathlib import PosixPath
from pydantic import (BaseModel,
                      confloat,
                      conint,
                      conlist,
                      constr,
                      FilePath,
                      root_validator,
                      validator)
from scipy.interpolate import interp1d
from typing import Literal, Optional

# local modules
from CHAP.utils.scanparsers import SMBMCAScanParser as ScanParser


class MCAScanDataConfig(BaseModel):
    """Class representing metadata required to locate raw MCA data for
    a single scan and construct a mask for it.

    :ivar spec_file: Path to the SPEC file containing the scan
    :ivar scan_number: Number of the scan in `spec_file`

    :ivar detector_name: name of the MCA used with the scan
    :ivar num_bins: number of channels on the MCA

    :ivar include_bin_ranges: list of MCA channel index ranges whose
        data should be included after applying a mask
    """
    spec_file: FilePath
    scan_number: conint(gt=0)

    detector_name: constr(strip_whitespace=True, min_length=1) = 'mca1'
    num_bins: Optional[conint(gt=0)]

    include_bin_ranges: Optional[
        conlist(
            min_items=1,
            item_type=conlist(
                item_type=conint(ge=0),
                min_items=2,
                max_items=2))] = None

    _scanparser: Optional[ScanParser]

    class Config:
        underscore_attrs_are_private = False

    @root_validator
    def validate_root(cls, values):
        """Validate the `values` dictionary. Fill in a value for
        `_scanparser` and `num_bins` (if the latter was not already
        provided)

        :param values: dictionary of field values to validate
        :type values: dict
        :return: the validated form of `values`
        :rtype: dict
        """
        values['_scanparser'] = ScanParser(values.get('spec_file'),
                                           values.get('scan_number'))
        if values.get('num_bins') is None:
            try:
                values['num_bins'] = values['_scanparser']\
                    .get_detector_num_bins(
                        values.get('detector_name'))
            except exc:
                raise ValueError('No value found for num_bins') from exc
        return values

    @property
    def scanparser(self):
        try:
            scanparser = self._scanparser
        except:
            scanparser = ScanParser(self.spec_file, self.scan_number)
            self._scanparser = scanparser
        return scanparser

    @validator('include_bin_ranges', each_item=True)
    def validate_include_bin_range(cls, value, values):
        """Ensure no bin ranges are outside the boundary of the detector"""
        num_bins = values.get('num_bins')
        if num_bins is not None:
            value[1] = min(value[1], num_bins)
        return value

    def mca_mask(self):
        """Get a boolean mask array to use on the MCA data.

        :return: boolean mask array
        :rtype: numpy.ndarray
        """
        mask = np.asarray([False] * self.num_bins)
        bin_indices = np.arange(self.num_bins)
        for min_, max_ in self.include_bin_ranges:
            _mask = np.logical_and(bin_indices > min_, bin_indices < max_)
            mask = np.logical_or(mask, _mask)
        return mask

    def mca_data(self, scan_step_index=None):
        """Get the array of MCA data collected by the scan.

        :return: MCA data
        :rtype: np.ndarray
        """
        if scan_step_index is None:
            data = self.scanparser.get_all_detector_data(self.detector_name)
        else:
            data = self.scanparser.get_detector_data(self.detector_name,
                                                     self.scan_step_index)
        return data

    def dict(self):
        """Return a representation of this configuration in a
        dictionary that is suitable for dumping to a YAML file.

        :return: dictionary representation of the configuration.
        :rtype: dict
        """
        d = super().dict()
        for k,v in d.items():
            if isinstance(v, PosixPath):
                d[k] = str(v)
        if '_scanparser' in d:
            del d['_scanparser']
        return d


class DiffractionVolumeLengthConfig(MCAScanDataConfig):
    """Class representing metadata required to perform a diffraction
    volume length calculation for an EDD setup using a steel-foil
    raster scan.

    :ivar measurement_mode: placeholder for recording whether the
        measured DVL value was obtained through the automated
        calculation or a manual selection.
    :type measurement_mode: Literal['manual', 'auto']
    :ivar sigma_to_dvl_factor: to measure the DVL, a gaussian is fit
        to a reduced from of the raster scan MCA data. This variable
        is a scalar that converts the standard deviation of the
        gaussian fit to the measured DVL.
    :type sigma_to_dvl_factor: Optional[Literal[1.75, 1., 2.]]
    :ivar dvl_measured: placeholder for the measured diffraction
        volume length before writing data to file.
    """
    measurement_mode: Optional[Literal['manual', 'auto']] = 'auto'
    sigma_to_dvl_factor: Optional[Literal[1.75, 1., 2.]] = 1.75
    dvl_measured: Optional[confloat(gt=0)] = None

    @property
    def motor_vals(self):
        """Return the list of values visited by the scanning motor
        over the course of the raster scan.

        :return: list of scanned motor values
        :rtype: np.ndarray
        """
        return self.scanparser.spec_scan_motor_vals[0]

    @property
    def motor_mne(self):
        """Return the mnenomic of the raster scan's motor"""
        return self.scanparser.spec_scan_motor_mnes[0]

    def dict(self):
        """If measurement_mode is 'manual', exclude
        sigma_to_dvl_factor from the dict representation.
        """
        d = super().dict()
        if self.measurement_mode == 'manual':
            del d['sigma_to_dvl_factor']
        return d


class MCACeriaCalibrationConfig(MCAScanDataConfig):
    """
    Class representing metadata required to perform a Ceria calibration for an
    MCA detector.

    :ivar scan_step_index: Index of the scan step to use for calibration,
        optional. If not specified, the calibration routine will be performed
        on the average of all MCA spectra for the scan.

    :ivar flux_file: csv file containing station beam energy in eV (column 0)
        and flux (column 1)

    :ivar max_energy_kev: maximum channel energy of the MCA in keV

    :ivar hexrd_h5_material_file: path to a HEXRD materials.h5 file containing
        an entry for the material properties.
    :ivar hexrd_h5_material_name: Name of the material entry in
        `hexrd_h5_material_file`, defaults to `'CeO2'`.
    :ivar lattice_parameter_angstrom: lattice spacing in angstrom to use for
        the cubic CeO2 crystal, defaults to `5.41153`.

    :ivar tth_max: detector rotation about hutch x axis, defaults to `90`.
    :ivar hkl_tth_tol: minimum resolvable difference in 2&theta between two
        unique HKL peaks, defaults to `0.15`.

    :ivar fit_hkls: list of unique HKL indices to fit peaks for in the
        calibration routine

    :ivar tth_initial_guess: initial guess for 2&theta
    :ivar slope_initial_guess: initial guess for detector channel energy
        correction linear slope, defaults to `1.0`.
    :ivar intercept_initial_guess: initial guess for detector channel energy
        correction y-intercept, defaults to `0.0`.

    :ivar tth_calibrated: calibrated value for 2&theta, defaults to None
    :ivar slope_calibrated: calibrated value for detector channel energy
        correction linear slope, defaults to `None`
    :ivar intercept_calibrated: calibrated value for detector channel energy
        correction y-intercept, defaluts to None

    :ivar max_iter: maximum number of iterations of the calibration routine,
        defaults to `10`.
    :ivar tune_tth_tol: stop iteratively tuning 2&theta when an iteration
        produces a change in the tuned value of 2&theta that is smaller than
        this value, defaults to `1e-8`.
    """
    scan_step_index: Optional[conint(ge=0)]

    flux_file: FilePath

    max_energy_kev: confloat(gt=0)

    hexrd_h5_material_file: FilePath
    hexrd_h5_material_name: constr(
        strip_whitespace=True, min_length=1) = 'CeO2'
    lattice_parameter_angstrom: confloat(gt=0) = 5.41153

    tth_max: confloat(gt=0, allow_inf_nan=False) = 90.0
    hkl_tth_tol: confloat(gt=0, allow_inf_nan=False) = 0.15

    fit_hkls: Optional[conlist(item_type=conint(ge=0), min_items=1)] = None

    tth_initial_guess: confloat(gt=0, le=tth_max, allow_inf_nan=False)
    slope_initial_guess: float = 1.0
    intercept_initial_guess: float = 0.0
    tth_calibrated: Optional[confloat(gt=0, allow_inf_nan=False)]
    slope_calibrated: Optional[confloat(allow_inf_nan=False)]
    intercept_calibrated: Optional[confloat(allow_inf_nan=False)]

    max_iter: conint(gt=0) = 10
    tune_tth_tol: confloat(ge=0) = 1e-8

    def mca_data(self):
        """Get the 1D array of MCA data to use for calibration.

        :return: MCA data
        :rtype: np.ndarray
        """
        if self.scan_step_index is None:
            data = super().mca_data()
            if self.scanparser.spec_scan_npts > 1:
                data = np.average(data, axis=1)
            else:
                data = data[0]
        else:
            data = super().mca_data(scan_step_index=self.scan_step_index)
        return data

    def flux_correction_interpolation_function(self):
        """
        Get an interpolation function to correct MCA data for relative energy
        flux of the incident beam.

        :return: energy flux correction interpolation function
        :rtype: scipy.interpolate._polyint._Interpolator1D
        """

        flux = np.loadtxt(self.flux_file)
        energies = flux[:,0]/1.e3
        relative_intensities = flux[:,1]/np.max(flux[:,1])
        interpolation_function = interp1d(energies, relative_intensities)
        return interpolation_function

    def material(self):
        """Get CeO2 as a `CHAP.utils.material.Material` object.

        :return: CeO2 material
        :rtype: CHAP.utils.material.Material
        """
        # local modules
        from CHAP.utils.material import Material

        material = Material(
            material_name=self.hexrd_h5_material_name,
            material_file=self.hexrd_h5_material_file,
            lattice_parameters_angstroms=self.lattice_parameter_angstrom)
        # The following kwargs will be needed if we allow the material to be
        # built using xrayutilities (for now, we only allow hexrd to make the
        # material):
        #   sgnum=225,
        #   atoms=['Ce4p', 'O2mdot'],
        #   pos=[(0.,0.,0.), (0.25,0.75,0.75)],
        #   enrgy=50000.)
        # Why do we need to specify an energy to get HKLs when using
        # xrayutilities?
        return material

    def unique_ds(self):
        """Get a list of unique HKLs and their lattice spacings

        :return: unique HKLs and their lattice spacings in angstroms
        :rtype: np.ndarray, np.ndarray
        """

        unique_hkls, unique_ds = self.material().get_ds_unique(
            tth_tol=self.hkl_tth_tol, tth_max=self.tth_max)

        return unique_hkls, unique_ds

    def fit_ds(self):
        """
        Get a list of HKLs and their lattice spacings that will be fit in the
        calibration routine

        :return: HKLs to fit and their lattice spacings in angstroms
        :rtype: np.ndarray, np.ndarray
        """

        unique_hkls, unique_ds = self.unique_ds()

        fit_hkls = np.array([unique_hkls[i] for i in self.fit_hkls])
        fit_ds = np.array([unique_ds[i] for i in self.fit_hkls])

        return fit_hkls, fit_ds
